@@ -49,8 +49,6 @@
 
 #define S(x)	(1 << (x))
 
-#define DNS 10
-
 struct gprs_ns2_vc_priv {
 	struct gprs_ns2_vc *nsvc;
 	/* how often the timer was triggered */
@@ -60,8 +58,6 @@ struct gprs_ns2_vc_priv {
 	bool initiator;
 	bool initiate_block;
 	bool initiate_reset;
-	/* if blocked by O&M/vty */
-	bool om_blocked;
 	/* if unitdata is forwarded to the user */
 	bool accept_unitdata;
 
@@ -117,8 +113,10 @@ enum gprs_ns2_vc_event {
 	GPRS_NS2_EV_RX_UNITDATA,
 
 	GPRS_NS2_EV_REQ_FORCE_UNCONFIGURED,	/* called via vty for tests */
+	GPRS_NS2_EV_REQ_OM_RESET,		/* vty cmd: reset */
 	GPRS_NS2_EV_REQ_OM_BLOCK,		/* vty cmd: block */
 	GPRS_NS2_EV_REQ_OM_UNBLOCK,		/* vty cmd: unblock*/
+	GPRS_NS2_EV_RX_BLOCK_FOREIGN,		/* received a BLOCK over another NSVC */
 };
 
 static const struct value_string ns2_vc_event_names[] = {
@@ -128,12 +126,14 @@ static const struct value_string ns2_vc_event_names[] = {
 	{ GPRS_NS2_EV_RX_UNBLOCK,		"RX-UNBLOCK" },
 	{ GPRS_NS2_EV_RX_UNBLOCK_ACK,		"RX-UNBLOCK_ACK" },
 	{ GPRS_NS2_EV_RX_BLOCK,			"RX-BLOCK" },
+	{ GPRS_NS2_EV_RX_BLOCK_FOREIGN,		"RX-BLOCK_FOREIGN" },
 	{ GPRS_NS2_EV_RX_BLOCK_ACK,		"RX-BLOCK_ACK" },
 	{ GPRS_NS2_EV_RX_ALIVE,			"RX-ALIVE" },
 	{ GPRS_NS2_EV_RX_ALIVE_ACK,		"RX-ALIVE_ACK" },
 	{ GPRS_NS2_EV_RX_STATUS,		"RX-STATUS" },
 	{ GPRS_NS2_EV_RX_UNITDATA,		"RX-UNITDATA" },
 	{ GPRS_NS2_EV_REQ_FORCE_UNCONFIGURED,	"REQ-FORCE_UNCONFIGURED" },
+	{ GPRS_NS2_EV_REQ_OM_RESET,		"REQ-O&M-RESET"},
 	{ GPRS_NS2_EV_REQ_OM_BLOCK,		"REQ-O&M-BLOCK"},
 	{ GPRS_NS2_EV_REQ_OM_UNBLOCK,		"REQ-O&M-UNBLOCK"},
 	{ 0, NULL }
@@ -145,26 +145,47 @@ static inline struct gprs_ns2_inst *ns_inst_from_fi(struct osmo_fsm_inst *fi)
 	return priv->nsvc->nse->nsi;
 }
 
-static void start_test_procedure(struct gprs_ns2_vc_priv *priv)
+/* Start the NS-TEST procedure, either with transmitting a tx_alive,
+ * (start_tx_alive==true) or with starting tns-test */
+static void start_test_procedure(struct osmo_fsm_inst *fi, bool start_tx_alive)
 {
+	struct gprs_ns2_vc_priv *priv = fi->priv;
 	struct gprs_ns2_inst *nsi = priv->nsvc->nse->nsi;
+	unsigned int tout_idx;
 
-	if (osmo_timer_pending(&priv->alive.timer))
-		return;
+	if (osmo_timer_pending(&priv->alive.timer)) {
+		if (start_tx_alive) {
+			if (priv->alive.mode == NS_TOUT_TNS_ALIVE)
+				return;
+		} else {
+			if (priv->alive.mode == NS_TOUT_TNS_TEST)
+				return;
+		}
+	}
 
-	priv->alive.mode = NS_TOUT_TNS_ALIVE;
 	priv->alive.N = 0;
 
-	osmo_clock_gettime(CLOCK_MONOTONIC, &priv->alive.timer_started);
-	ns2_tx_alive(priv->nsvc);
-	osmo_timer_schedule(&priv->alive.timer, nsi->timeout[NS_TOUT_TNS_ALIVE], 0);
+	if (start_tx_alive) {
+		priv->alive.mode = NS_TOUT_TNS_ALIVE;
+		osmo_clock_gettime(CLOCK_MONOTONIC, &priv->alive.timer_started);
+		ns2_tx_alive(priv->nsvc);
+		tout_idx = NS_TOUT_TNS_ALIVE;
+	} else {
+		priv->alive.mode = NS_TOUT_TNS_TEST;
+		tout_idx = NS_TOUT_TNS_TEST;
+	}
+	LOGPFSML(fi, LOGL_DEBUG, "Starting Tns-%s of %u seconds\n",
+		 tout_idx == NS_TOUT_TNS_ALIVE ? "alive" : "test", nsi->timeout[tout_idx]);
+	osmo_timer_schedule(&priv->alive.timer, nsi->timeout[tout_idx], 0);
 }
 
 static void stop_test_procedure(struct gprs_ns2_vc_priv *priv)
 {
+	osmo_stat_item_set(osmo_stat_item_group_get_item(priv->nsvc->statg, NS_STAT_ALIVE_DELAY), 0);
 	osmo_timer_del(&priv->alive.timer);
 }
 
+/* how many milliseconds have expired since the last alive timer start? */
 static int alive_timer_elapsed_ms(struct gprs_ns2_vc_priv *priv)
 {
 	struct timespec now, elapsed;
@@ -176,6 +197,7 @@ static int alive_timer_elapsed_ms(struct gprs_ns2_vc_priv *priv)
 	return elapsed.tv_sec * 1000 + (elapsed.tv_nsec / 1000000);
 }
 
+/* we just received a NS-ALIVE-ACK; re-schedule after Tns-test */
 static void recv_test_procedure(struct osmo_fsm_inst *fi)
 {
 	struct gprs_ns2_vc_priv *priv = fi->priv;
@@ -188,7 +210,7 @@ static void recv_test_procedure(struct osmo_fsm_inst *fi)
 
 	priv->alive.mode = NS_TOUT_TNS_TEST;
 	osmo_timer_schedule(&priv->alive.timer, nsi->timeout[NS_TOUT_TNS_TEST], 0);
-	osmo_stat_item_set(nsvc->statg->items[NS_STAT_ALIVE_DELAY],
+	osmo_stat_item_set(osmo_stat_item_group_get_item(nsvc->statg, NS_STAT_ALIVE_DELAY),
 		alive_timer_elapsed_ms(priv));
 }
 
@@ -208,7 +230,7 @@ static void alive_timeout_handler(void *data)
 		osmo_timer_schedule(&priv->alive.timer, nsi->timeout[NS_TOUT_TNS_ALIVE], 0);
 		break;
 	case NS_TOUT_TNS_ALIVE:
-		rate_ctr_inc(&priv->nsvc->ctrg->ctr[NS_CTR_LOST_ALIVE]);
+		RATE_CTR_INC_NS(priv->nsvc, NS_CTR_LOST_ALIVE);
 		priv->alive.N++;
 
 		if (priv->alive.N <= nsi->timeout[NS_TOUT_TNS_ALIVE_RETRIES]) {
@@ -232,7 +254,10 @@ static void alive_timeout_handler(void *data)
 
 static void ns2_st_unconfigured_onenter(struct osmo_fsm_inst *fi, uint32_t old_state)
 {
+	struct gprs_ns2_vc_priv *priv = fi->priv;
+
 	stop_test_procedure(fi->priv);
+	ns2_nse_notify_unblocked(priv->nsvc, false);
 }
 
 static void ns2_st_unconfigured(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -241,19 +266,24 @@ static void ns2_st_unconfigured(struct osmo_fsm_inst *fi, uint32_t event, void *
 	struct gprs_ns2_inst *nsi = priv->nsvc->nse->nsi;
 
 	priv->initiate_reset = priv->initiate_block = priv->initiator;
-	priv->om_blocked = false;
+	priv->nsvc->om_blocked = false;
 
 	switch (event) {
 	case GPRS_NS2_EV_REQ_START:
 		switch (priv->nsvc->mode) {
 		case GPRS_NS2_VC_MODE_ALIVE:
-			osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RECOVERING, nsi->timeout[NS_TOUT_TNS_ALIVE], NS_TOUT_TNS_ALIVE);
+			if (priv->nsvc->nse->dialect == GPRS_NS2_DIALECT_SNS) {
+				/* In IP-SNS, the NS-VC are assumed initially alive, until the alive
+				* procedure should fail at some future point */
+				osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_UNBLOCKED, 0, 0);
+			} else {
+				osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RECOVERING, nsi->timeout[NS_TOUT_TNS_ALIVE], NS_TOUT_TNS_ALIVE);
+			}
 			break;
 		case GPRS_NS2_VC_MODE_BLOCKRESET:
 			osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RESET, nsi->timeout[NS_TOUT_TNS_RESET], NS_TOUT_TNS_RESET);
 			break;
 		}
-
 		break;
 	default:
 		OSMO_ASSERT(0);
@@ -309,49 +339,57 @@ static void ns2_st_blocked_onenter(struct osmo_fsm_inst *fi, uint32_t old_state)
 
 	if (old_state != GPRS_NS2_ST_BLOCKED) {
 		priv->N = 0;
-		rate_ctr_inc(&priv->nsvc->ctrg->ctr[NS_CTR_BLOCKED]);
+		RATE_CTR_INC_NS(priv->nsvc, NS_CTR_BLOCKED);
 	}
 
-	if (priv->om_blocked) {
+	ns2_nse_notify_unblocked(priv->nsvc, false);
+	if (priv->nsvc->om_blocked) {
 		/* we are already blocked after a RESET */
 		if (old_state == GPRS_NS2_ST_RESET) {
 			osmo_timer_del(&fi->timer);
 		} else {
-			ns2_tx_block(priv->nsvc, NS_CAUSE_OM_INTERVENTION);
+			ns2_tx_block(priv->nsvc, NS_CAUSE_OM_INTERVENTION, NULL);
 		}
 	} else if (priv->initiate_block) {
 		ns2_tx_unblock(priv->nsvc);
 	}
 
-	start_test_procedure(priv);
+	start_test_procedure(fi, true);
 }
 
 static void ns2_st_blocked(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct gprs_ns2_vc_priv *priv = fi->priv;
 
-	if (priv->om_blocked) {
+	if (priv->nsvc->om_blocked) {
 		switch (event) {
 		case GPRS_NS2_EV_RX_BLOCK_ACK:
 			priv->accept_unitdata = false;
 			osmo_timer_del(&fi->timer);
 			break;
 		case GPRS_NS2_EV_RX_BLOCK:
+			ns2_tx_block_ack(priv->nsvc, NULL);
+			/* fall through */
+		case GPRS_NS2_EV_RX_BLOCK_FOREIGN:
+			/* the BLOCK ACK for foreign BLOCK PDUs (rx over another nsvc) will be sent
+			 * from the receiving nsvc */
 			priv->accept_unitdata = false;
-			ns2_tx_block_ack(priv->nsvc);
 			osmo_timer_del(&fi->timer);
 			break;
 		case GPRS_NS2_EV_RX_UNBLOCK:
 			priv->accept_unitdata = false;
-			ns2_tx_block(priv->nsvc, NS_CAUSE_OM_INTERVENTION);
+			ns2_tx_block(priv->nsvc, NS_CAUSE_OM_INTERVENTION, NULL);
 			osmo_timer_add(&fi->timer);
 			break;
 		}
 	} else if (priv->initiate_block) {
 		switch (event) {
+		case GPRS_NS2_EV_RX_BLOCK_FOREIGN:
+			/* the block ack will be sent by the rx NSVC */
+			break;
 		case GPRS_NS2_EV_RX_BLOCK:
 			/* TODO: BLOCK is a UNBLOCK_NACK */
-			ns2_tx_block_ack(priv->nsvc);
+			ns2_tx_block_ack(priv->nsvc, NULL);
 			break;
 		case GPRS_NS2_EV_RX_UNBLOCK:
 			ns2_tx_unblock_ack(priv->nsvc);
@@ -365,8 +403,11 @@ static void ns2_st_blocked(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 	} else {
 		/* we are on the receiving end. The initiator who sent RESET is responsible to UNBLOCK! */
 		switch (event) {
+		case GPRS_NS2_EV_RX_BLOCK_FOREIGN:
+			/* the block ack will be sent by the rx NSVC */
+			break;
 		case GPRS_NS2_EV_RX_BLOCK:
-			ns2_tx_block_ack(priv->nsvc);
+			ns2_tx_block_ack(priv->nsvc, NULL);
 			break;
 		case GPRS_NS2_EV_RX_UNBLOCK:
 			ns2_tx_unblock_ack(priv->nsvc);
@@ -383,12 +424,20 @@ static void ns2_st_unblocked_on_enter(struct osmo_fsm_inst *fi, uint32_t old_sta
 	struct gprs_ns2_vc *nsvc = priv->nsvc;
 	struct gprs_ns2_nse *nse = nsvc->nse;
 
-	if (old_state != GPRS_NS2_ST_UNBLOCKED)
-		rate_ctr_inc(&nsvc->ctrg->ctr[NS_CTR_UNBLOCKED]);
+	if (old_state != GPRS_NS2_ST_UNBLOCKED) {
+		RATE_CTR_INC_NS(nsvc, NS_CTR_UNBLOCKED);
+		osmo_clock_gettime(CLOCK_MONOTONIC, &nsvc->ts_alive_change);
+	}
 
 	priv->accept_unitdata = true;
 	ns2_nse_notify_unblocked(nsvc, true);
 	ns2_prim_status_ind(nse, nsvc, 0, GPRS_NS2_AFF_CAUSE_VC_RECOVERY);
+
+	/* the closest interpretation of the spec would start Tns-test here first,
+	 * and only send a NS-ALIVE after Tns-test has expired (i.e. setting the
+	 * second argument to 'false'.  However, being quick in detecting unavailability
+	 * of a NS-VC seems like a good idea */
+	start_test_procedure(fi, true);
 }
 
 static void ns2_st_unblocked(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -400,8 +449,13 @@ static void ns2_st_unblocked(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 		ns2_tx_unblock_ack(priv->nsvc);
 		break;
 	case GPRS_NS2_EV_RX_BLOCK:
+		ns2_tx_block_ack(priv->nsvc, NULL);
+		/* fall through */
+	case GPRS_NS2_EV_RX_BLOCK_FOREIGN:
+		/* the BLOCK ACK for foreign BLOCK PDUs (rx over another nsvc) will be sent
+		 * from the receiving nsvc */
 		priv->initiate_block = false;
-		ns2_tx_block_ack(priv->nsvc);
+		priv->accept_unitdata = false;
 		osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_BLOCKED,
 					0, 2);
 		break;
@@ -428,19 +482,16 @@ static void ns2_st_alive_onenter(struct osmo_fsm_inst *fi, uint32_t old_state)
 	if (old_state != GPRS_NS2_ST_RECOVERING)
 		priv->N = 0;
 
-	ns2_tx_alive(priv->nsvc);
+	start_test_procedure(fi, true);
 	ns2_nse_notify_unblocked(priv->nsvc, false);
 }
 
-static void ns2_st_alive_onleave(struct osmo_fsm_inst *fi, uint32_t next_state)
-{
-	start_test_procedure(fi->priv);
-}
-
 static const struct osmo_fsm_state ns2_vc_states[] = {
 	[GPRS_NS2_ST_UNCONFIGURED] = {
 		.in_event_mask = S(GPRS_NS2_EV_REQ_START),
-		.out_state_mask = S(GPRS_NS2_ST_RESET) | S(GPRS_NS2_ST_RECOVERING),
+		.out_state_mask = S(GPRS_NS2_ST_RESET) |
+				  S(GPRS_NS2_ST_RECOVERING) |
+				  S(GPRS_NS2_ST_UNBLOCKED),
 		.name = "UNCONFIGURED",
 		.action = ns2_st_unconfigured,
 		.onenter = ns2_st_unconfigured_onenter,
@@ -456,7 +507,8 @@ static const struct osmo_fsm_state ns2_vc_states[] = {
 	},
 	[GPRS_NS2_ST_BLOCKED] = {
 		.in_event_mask = S(GPRS_NS2_EV_RX_BLOCK) | S(GPRS_NS2_EV_RX_BLOCK_ACK) |
-		S(GPRS_NS2_EV_RX_UNBLOCK) | S(GPRS_NS2_EV_RX_UNBLOCK_ACK),
+				 S(GPRS_NS2_EV_RX_UNBLOCK) | S(GPRS_NS2_EV_RX_UNBLOCK_ACK) |
+				 S(GPRS_NS2_EV_RX_BLOCK_FOREIGN),
 		.out_state_mask = S(GPRS_NS2_ST_RESET) |
 				  S(GPRS_NS2_ST_UNBLOCKED) |
 				  S(GPRS_NS2_ST_BLOCKED) |
@@ -467,7 +519,7 @@ static const struct osmo_fsm_state ns2_vc_states[] = {
 	},
 	[GPRS_NS2_ST_UNBLOCKED] = {
 		.in_event_mask = S(GPRS_NS2_EV_RX_BLOCK) | S(GPRS_NS2_EV_RX_UNBLOCK_ACK) |
-				 S(GPRS_NS2_EV_RX_UNBLOCK),
+				 S(GPRS_NS2_EV_RX_UNBLOCK) | S(GPRS_NS2_EV_RX_BLOCK_FOREIGN),
 		.out_state_mask = S(GPRS_NS2_ST_RESET) | S(GPRS_NS2_ST_RECOVERING) |
 				  S(GPRS_NS2_ST_BLOCKED) |
 				  S(GPRS_NS2_ST_UNCONFIGURED),
@@ -485,7 +537,6 @@ static const struct osmo_fsm_state ns2_vc_states[] = {
 		.name = "RECOVERING",
 		.action = ns2_st_alive,
 		.onenter = ns2_st_alive_onenter,
-		.onleave = ns2_st_alive_onleave,
 	},
 };
 
@@ -497,7 +548,7 @@ static int ns2_vc_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	switch (fi->state) {
 	case GPRS_NS2_ST_RESET:
 		if (priv->initiate_reset) {
-			rate_ctr_inc(&priv->nsvc->ctrg->ctr[NS_CTR_LOST_RESET]);
+			RATE_CTR_INC_NS(priv->nsvc, NS_CTR_LOST_RESET);
 			priv->N++;
 			if (priv->N <= nsi->timeout[NS_TOUT_TNS_RESET_RETRIES]) {
 				osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RESET, nsi->timeout[NS_TOUT_TNS_RESET], 0);
@@ -510,7 +561,7 @@ static int ns2_vc_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	case GPRS_NS2_ST_BLOCKED:
 		if (priv->initiate_block) {
 			priv->N++;
-			if (priv->om_blocked) {
+			if (priv->nsvc->om_blocked) {
 				if (priv->N <= nsi->timeout[NS_TOUT_TNS_BLOCK_RETRIES]) {
 					osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_BLOCKED, nsi->timeout[NS_TOUT_TNS_BLOCK], 0);
 				} else {
@@ -582,9 +633,20 @@ static void ns2_vc_fsm_allstate_action(struct osmo_fsm_inst *fi,
 {
 	struct gprs_ns2_vc_priv *priv = fi->priv;
 	struct gprs_ns2_inst *nsi = ns_inst_from_fi(fi);
+	struct tlv_parsed *tp;
 	struct msgb *msg = data;
+	uint8_t cause;
 
 	switch (event) {
+	case GPRS_NS2_EV_REQ_OM_RESET:
+		if (priv->nsvc->mode != GPRS_NS2_VC_MODE_BLOCKRESET)
+			break;
+		/* move the FSM into reset */
+		if (fi->state != GPRS_NS2_ST_RESET) {
+			priv->initiate_reset = true;
+			osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RESET, nsi->timeout[NS_TOUT_TNS_RESET], NS_TOUT_TNS_RESET);
+		}
+		break;
 	case GPRS_NS2_EV_RX_RESET:
 		if (priv->nsvc->mode != GPRS_NS2_VC_MODE_BLOCKRESET)
 			break;
@@ -620,6 +682,8 @@ static void ns2_vc_fsm_allstate_action(struct osmo_fsm_inst *fi,
 		 * the msg, the upper layer has to do it.
 		 * Otherwise the msg must be freed.
 		 */
+
+		LOG_NS_DATA(priv->nsvc, "Rx", NS_PDUT_UNITDATA, LOGL_INFO, "\n");
 		switch (fi->state) {
 		case GPRS_NS2_ST_BLOCKED:
 			/* 7.2.1: the BLOCKED_ACK might be lost */
@@ -630,7 +694,7 @@ static void ns2_vc_fsm_allstate_action(struct osmo_fsm_inst *fi,
 
 			ns2_tx_status(priv->nsvc,
 				      NS_CAUSE_NSVC_BLOCKED,
-				      0, msg);
+				      0, msg, NULL);
 			break;
 		/* ALIVE can receive UNITDATA if the ALIVE_ACK is lost */
 		case GPRS_NS2_ST_RECOVERING:
@@ -651,17 +715,38 @@ static void ns2_vc_fsm_allstate_action(struct osmo_fsm_inst *fi,
 	case GPRS_NS2_EV_REQ_OM_BLOCK:
 		/* vty cmd: block */
 		priv->initiate_block = true;
-		priv->om_blocked = true;
+		priv->nsvc->om_blocked = true;
 		osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_BLOCKED, nsi->timeout[NS_TOUT_TNS_BLOCK], 0);
 		break;
 	case GPRS_NS2_EV_REQ_OM_UNBLOCK:
 		/* vty cmd: unblock*/
-		if (!priv->om_blocked)
+		if (!priv->nsvc->om_blocked)
 			return;
-		priv->om_blocked = false;
+		priv->nsvc->om_blocked = false;
 		if (fi->state == GPRS_NS2_ST_BLOCKED)
 			osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_BLOCKED, nsi->timeout[NS_TOUT_TNS_BLOCK], 0);
 		break;
+	case GPRS_NS2_EV_RX_STATUS:
+		tp = data;
+		cause = tlvp_val8(tp, NS_IE_CAUSE, 0);
+		switch (cause) {
+		case NS_CAUSE_NSVC_BLOCKED:
+			if (fi->state != GPRS_NS2_ST_BLOCKED) {
+				LOG_NS_SIGNAL(priv->nsvc, "Rx", NS_PDUT_STATUS, LOGL_ERROR, ": remote side reported blocked state.\n");
+				priv->initiate_block = false;
+				priv->accept_unitdata = false;
+				osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_BLOCKED, nsi->timeout[NS_TOUT_TNS_BLOCK], 0);
+			}
+			break;
+		case NS_CAUSE_NSVC_UNKNOWN:
+			if (fi->state != GPRS_NS2_ST_RESET && fi->state != GPRS_NS2_ST_UNCONFIGURED) {
+				LOG_NS_SIGNAL(priv->nsvc, "Rx", NS_PDUT_STATUS, LOGL_ERROR, ": remote side reported unknown nsvc.\n");
+				osmo_fsm_inst_state_chg(fi, GPRS_NS2_ST_RESET, nsi->timeout[NS_TOUT_TNS_RESET], 0);
+			}
+			break;
+		}
+
+		break;
 	}
 }
 
@@ -681,13 +766,14 @@ static struct osmo_fsm ns2_vc_fsm = {
 			       S(GPRS_NS2_EV_RX_RESET) |
 			       S(GPRS_NS2_EV_RX_ALIVE) |
 			       S(GPRS_NS2_EV_RX_ALIVE_ACK) |
+			       S(GPRS_NS2_EV_RX_STATUS) |
 			       S(GPRS_NS2_EV_REQ_FORCE_UNCONFIGURED) |
+			       S(GPRS_NS2_EV_REQ_OM_RESET) |
 			       S(GPRS_NS2_EV_REQ_OM_BLOCK) |
 			       S(GPRS_NS2_EV_REQ_OM_UNBLOCK),
 	.allstate_action = ns2_vc_fsm_allstate_action,
 	.cleanup = ns2_vc_fsm_clean,
 	.timer_cb = ns2_vc_fsm_timer_cb,
-	/* .log_subsys = DNS, "is not constant" */
 	.event_names = ns2_vc_event_names,
 	.pre_term = NULL,
 	.log_subsys = DLNS,
@@ -745,6 +831,10 @@ int ns2_vc_force_unconfigured(struct gprs_ns2_vc *nsvc)
  *  \return 0 on success; negative on error */
 int ns2_vc_block(struct gprs_ns2_vc *nsvc)
 {
+	struct gprs_ns2_vc_priv *priv = nsvc->fi->priv;
+	if (priv->nsvc->om_blocked)
+		return -EALREADY;
+
 	return osmo_fsm_inst_dispatch(nsvc->fi, GPRS_NS2_EV_REQ_OM_BLOCK, NULL);
 }
 
@@ -753,9 +843,21 @@ int ns2_vc_block(struct gprs_ns2_vc *nsvc)
  *  \return 0 on success; negative on error */
 int ns2_vc_unblock(struct gprs_ns2_vc *nsvc)
 {
+	struct gprs_ns2_vc_priv *priv = nsvc->fi->priv;
+	if (!priv->nsvc->om_blocked)
+		return -EALREADY;
+
 	return osmo_fsm_inst_dispatch(nsvc->fi, GPRS_NS2_EV_REQ_OM_UNBLOCK, NULL);
 }
 
+/*! Reset a NS-VC.
+ *  \param nsvc the virtual circuit
+ *  \return 0 on success; negative on error */
+int ns2_vc_reset(struct gprs_ns2_vc *nsvc)
+{
+	return osmo_fsm_inst_dispatch(nsvc->fi, GPRS_NS2_EV_REQ_OM_RESET, NULL);
+}
+
 /*! entry point for messages from the driver/VL
  *  \param nsvc virtual circuit on which the message was received
  *  \param msg message that was received
@@ -764,6 +866,7 @@ int ns2_vc_unblock(struct gprs_ns2_vc *nsvc)
 int ns2_vc_rx(struct gprs_ns2_vc *nsvc, struct msgb *msg, struct tlv_parsed *tp)
 {
 	struct gprs_ns_hdr *nsh = (struct gprs_ns_hdr *) msg->l2h;
+	struct gprs_ns2_vc *target_nsvc = nsvc;
 	struct osmo_fsm_inst *fi = nsvc->fi;
 	int rc = 0;
 	uint8_t cause;
@@ -774,8 +877,9 @@ int ns2_vc_rx(struct gprs_ns2_vc *nsvc, struct msgb *msg, struct tlv_parsed *tp)
 	/* TODO: handle BLOCK/UNBLOCK/ALIVE with different VCI */
 
 	if (ns2_validate(nsvc, nsh->pdu_type, msg, tp, &cause)) {
+		/* don't answer on a STATUS with a STATUS */
 		if (nsh->pdu_type != NS_PDUT_STATUS) {
-			rc = ns2_tx_status(nsvc, cause, 0, msg);
+			rc = ns2_tx_status(nsvc, cause, 0, msg, NULL);
 			goto out;
 		}
 	}
@@ -787,22 +891,41 @@ int ns2_vc_rx(struct gprs_ns2_vc *nsvc, struct msgb *msg, struct tlv_parsed *tp)
 			if (nsh->pdu_type == NS_PDUT_RESET)
 				ns2_tx_reset_ack(nsvc);
 
-			LOGNSVC(nsvc, LOGL_ERROR, "Rx %s with wrong NSEI=%05u. Ignoring PDU.\n",
-				get_value_string(gprs_ns_pdu_strings, nsh->pdu_type), nsei);
+			LOG_NS_SIGNAL(nsvc, "Rx", nsh->pdu_type, LOGL_ERROR, " with wrong NSEI (exp: %05u, got %05u). Ignoring PDU.\n", nsvc->nse->nsei, nsei);
 			goto out;
 		}
 	}
 
-	if (nsvc->nsvci_is_valid && TLVP_PRESENT(tp,  NS_IE_VCI)) {
+	if (nsvc->nsvci_is_valid && TLVP_PRESENT(tp, NS_IE_VCI)) {
 		nsvci = tlvp_val16be(tp, NS_IE_VCI);
 		if (nsvci != nsvc->nsvci) {
 			/* 48.016 § 7.3.1 send RESET_ACK to wrong NSVCI + ignore */
-			if (nsh->pdu_type == NS_PDUT_RESET)
+			if (nsh->pdu_type == NS_PDUT_RESET) {
 				ns2_tx_reset_ack(nsvc);
+				LOG_NS_SIGNAL(nsvc, "Rx", nsh->pdu_type, LOGL_ERROR, " with wrong NSVCI (exp: %05u, got %05u). Ignoring PDU.\n", nsvc->nsvci, nsvci);
+				goto out;
+			} else if (nsh->pdu_type == NS_PDUT_BLOCK || nsh->pdu_type == NS_PDUT_STATUS) {
+				/* this is a PDU received over a NSVC and reports a status/block for another NSVC */
+				target_nsvc = gprs_ns2_nsvc_by_nsvci(nsvc->nse->nsi,  nsvci);
+				if (!target_nsvc) {
+					LOGPFSML(fi, LOGL_ERROR, "Received a %s PDU for unknown NSVC (NSVCI %d)\n",
+						 get_value_string(gprs_ns_pdu_strings, nsh->pdu_type), nsvci);
+					if (nsh->pdu_type == NS_PDUT_BLOCK)
+						ns2_tx_status(nsvc, NS_CAUSE_NSVC_UNKNOWN, 0, msg, &nsvci);
+					goto out;
+				}
 
-			LOGNSVC(nsvc, LOGL_ERROR, "Rx %s with wrong NSVCI=%05u. Ignoring PDU.\n",
-				get_value_string(gprs_ns_pdu_strings, nsh->pdu_type), nsvci);
-			goto out;
+				if (target_nsvc->nse != nsvc->nse) {
+					LOGPFSML(fi, LOGL_ERROR, "Received a %s PDU for a NSVC (NSVCI %d) but it belongs to a different NSE!\n",
+						 get_value_string(gprs_ns_pdu_strings, nsh->pdu_type), nsvci);
+					goto out;
+				}
+
+				/* the status/block will be passed to the nsvc/target nsvc in the switch */
+			} else {
+				LOG_NS_SIGNAL(nsvc, "Rx", nsh->pdu_type, LOGL_ERROR, " with wrong NSVCI=%05u. Ignoring PDU.\n", nsvci);
+				goto out;
+			}
 		}
 	}
 
@@ -814,7 +937,12 @@ int ns2_vc_rx(struct gprs_ns2_vc *nsvc, struct msgb *msg, struct tlv_parsed *tp)
 		osmo_fsm_inst_dispatch(fi, GPRS_NS2_EV_RX_RESET_ACK, tp);
 		break;
 	case NS_PDUT_BLOCK:
-		osmo_fsm_inst_dispatch(fi, GPRS_NS2_EV_RX_BLOCK, tp);
+		if (target_nsvc != nsvc) {
+			osmo_fsm_inst_dispatch(target_nsvc->fi, GPRS_NS2_EV_RX_BLOCK_FOREIGN, tp);
+			ns2_tx_block_ack(nsvc, &nsvci);
+		} else {
+			osmo_fsm_inst_dispatch(fi, GPRS_NS2_EV_RX_BLOCK, tp);
+		}
 		break;
 	case NS_PDUT_BLOCK_ACK:
 		osmo_fsm_inst_dispatch(fi, GPRS_NS2_EV_RX_BLOCK_ACK, tp);
@@ -835,10 +963,14 @@ int ns2_vc_rx(struct gprs_ns2_vc *nsvc, struct msgb *msg, struct tlv_parsed *tp)
 		/* UNITDATA have to free msg because it might send the msg layer upwards */
 		osmo_fsm_inst_dispatch(fi, GPRS_NS2_EV_RX_UNITDATA, msg);
 		return 0;
+	case NS_PDUT_STATUS:
+		osmo_fsm_inst_dispatch(target_nsvc->fi, GPRS_NS2_EV_RX_STATUS, tp);
+		break;
 	default:
 		LOGPFSML(fi, LOGL_ERROR, "NSEI=%u Rx unknown NS PDU type %s\n", nsvc->nse->nsei,
 			 get_value_string(gprs_ns_pdu_strings, nsh->pdu_type));
-		return -EINVAL;
+		rc = -EINVAL;
+		break;
 	}
 
 out:
